/****************************************************************************
*                                                                           *
* asm-xml.h                                                                 *
*                                                                           *
* Copyright (c) 2007 Marc Kerbiquet                                         *
*                                                                           *
****************************************************************************/

#ifdef WIN32
  #define ACC __cdecl
#else
  #define ACC
#endif

#ifdef __cplusplus
extern "C" {
#endif

//-----------------------------------------------------------------------------
// Constants
//-----------------------------------------------------------------------------

#define RC_OK                       0 // everything is ok
#define RC_MEMORY                   1 // out of memory
#define RC_EMPTY_NAME               2 // name empty or not defined
#define RC_ATTR_DEFINED             3 // attribute already defined
#define RC_INVALID_ENTITY_REFERENCE 4 // must be amp, quot, lt, gt, or apos
#define RC_UNEXPECTED_END           5 // found last char too early
#define RC_INVALID_CHAR             6 // wrong char
#define RC_OVERFLOW                 7 // number to big in char reference
#define RC_NO_START_TAG             8 // xml does not start with a tag
#define RC_TAG_MISMATCH             9 // invalid close tag
#define RC_INVALID_TAG             10 // invalid root element

//-----------------------------------------------------------------------------
// Structures
//-----------------------------------------------------------------------------

typedef struct AXElement      AXElement       ;
typedef struct AXAttribute    AXAttribute     ;
typedef struct AXElementClass AXElementClass  ;
typedef struct AXParseContext AXParseContext  ;
typedef struct AXClassContext AXClassContext  ;

struct AXElementClass
{
  char*           name          ; // Name of the element (not zero terminated)
  char*           nameLimit     ; // End of the name of the element
  unsigned int    size          ; // size in bytes of an element of this class
  unsigned int    propertyCount ; // number of attributes and text elements
  unsigned int    childCount    ; // number of child classes
  int*            attributes    ; // (internal) attribute map
  int*            elements      ; // (internal) element map
  AXElementClass* children[1]   ; // The list of child classes.
                                  // The order is the one defined in the class
                                  // definition file.
                                  // No bound checking in C
};

struct AXParseContext
{
  void*           base          ;
  void*           limit         ;
  void*           chunks        ;
  int             chunkSize     ;
  int             errorCode     ;
  const char*     source        ;
  const char*     current       ;
  int             line          ;
  int             column        ;
  AXElement*      root          ;
};

struct AXClassContext
{
  void*           base          ;
  void*           limit         ;
  void*           chunks        ;
  int             chunkSize     ;
  int             errorCode     ;
};

struct AXAttribute
{
  char*           begin         ; // the value (not zero terminated)
  char*           limit         ; // the end of the value
};

struct AXElement
{
  AXElementClass* type          ; // the class of the element
  AXElement*      nextSibling   ; // the next sibling element
  AXElement*      firstChild    ; // the first child element
  AXElement*      lastChild     ; // the last child element
  AXAttribute     attributes[1] ; // the array of attributes - there is 
                                  // no bound checking in C
};

//-----------------------------------------------------------------------------
// Functions
//-----------------------------------------------------------------------------

extern 
void            ACC ax_initialize            (void*             mallocFun, 
                                              void*             freeFun);
extern 
int             ACC ax_initializeParser      (AXParseContext*   context, 
                                              unsigned int      chunkSize);
extern 
int             ACC ax_releaseParser         (AXParseContext*   context);
extern 
AXElement*      ACC ax_parse                 (AXParseContext*   context, 
                                              const char*       source,
                                              AXElementClass*   type);
extern 
int             ACC ax_initializeClassParser (AXClassContext*   context);
extern 
int             ACC ax_releaseClassParser    (AXClassContext*   context);
extern 
AXElementClass* ACC ax_classFromElement      (AXElement*        e,
                                              AXClassContext*   context);
extern 
AXElementClass* ACC ax_classFromString       (const char*       source, 
                                              AXClassContext*   context);

#ifdef __cplusplus
}
#endif
