;[c]asm-xml.asm - Asm XML Parser
;[c]
;[c]Copyright (c) 2007, Marc Kerbiquet
;[c]All rights reserved.
;[c]
;[c]Redistribution and use in source and binary forms, with or without
;[c]modification, are permitted provided that the following conditions are met:
;[c]
;[c]1. Redistributions of source code must retain the above copyright notice,
;[c]   this list of conditions and the following disclaimer.
;[c]2. Redistributions in binary form must reproduce the above copyright notice,
;[c]   this list of conditions and the following disclaimer in the documentation
;[c]   and/or other materials provided with the distribution.
;[c]3. Neither the name of AsmXml nor the names of its contributors may be
;[c]   used to endorse or promote products derived from this software without
;[c]   specific prior written permission.
;[c]
;[c]THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
;[c]AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
;[c]IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
;[c]DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
;[c]FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
;[c]DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
;[c]SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
;[c]CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
;[c]OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
;[c]OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
;[c]
;[c]
;[c]Open this file with Code Browser (http://code-browser.sourceforge.net/)
;[c]to view this long flat file as a structured text file.
;[c]
;[c]This is source code for  fasm.
;[c]
;[c]Conventions:
;[c]Preserved registers: ebx esi edi ebp
;[c]Scratch registers: eax ecx edx
;[c](same as Windows/Linux/BSD)
;[c]
;[of]:Macros
macro struct name
{
	virtual at 0
	label name
}

macro ends
{
.sizeof	rd	0
	end virtual
}
;[cf]
;[of]:Definitions
;[c]
		struct	MemoryContext
.base		rd	1	; memory buffer
.limit		rd	1	; limit of memory buffer
.chunks		rd	1	; list of chunks
.chunkSize	rd	1	; default size of chunks
		ends		
;[c]
		struct	ParseContext
.base		rd	1	; memory buffer
.limit		rd	1	; limit of memory buffer
.chunks		rd	1	; list of chunks
.chunkSize	rd	1	; default size of chunks
.errorCode	rd	1
.source		rd	1	; address of the first char
.current	rd	1	; address of the last read char
.line		rd	1	; line number of the last read char
.column		rd	1	; column of the last read char
.root		rd	1	; root element
		ends		
;[c]
		struct	CreateClassContext
.base		rd	1	; memory buffer
.limit		rd	1	; limit of memory buffer
.chunks		rd	1	; list of chunks
.chunkSize	rd	1	; default size of chunks
.errorCode	rd	1
		ends
;[c]
		struct	ElementClass
.name		rd	1	; name of the element
.nameLimit	rd	1	; limit of the name
.size		rd	1	; size to store the element
.propertyCount	rd	1	; number of attributes + text elements
.childCount	rd	1	; number of child classes
.attributes	rd	1	; list of attributes
.elements	rd	1	; list of elements
.children	rd	0	; list of child classes
		ends
;[c]
		struct TextElementClass
.offset		rd	1	; Offset of attribute in element
				; (the text element is stored as an attribute)
.begin		rd	1	; element name (begin)
.limit		rd	1	; element name (last char+1)
		ends
;[c]
		struct	Element
.type		rd	1
.nextSibling	rd	1
.firstChild	rd	1
.lastChild	rd	1
		ends
;[c]
;[c]
		struct	Attribute
.begin		rd	1
.limit		rd	1
		ends
;[c]
;[c]
EC_TEXT		equ	1
EC_TEXT_COLL	equ	2

;[c]
;[c]Error Codes
;[c]
RC_OK				equ	0	; everything is ok
RC_MEMORY			equ	1	; out of memory
RC_EMPTY_NAME			equ	2	; name empty or not defined
RC_ATTR_DEFINED			equ	3	; attribute already defined

RC_INVALID_ENTITY_REFERENCE	equ	4
RC_UNEXPECTED_END		equ	5
RC_INVALID_CHAR			equ	6
RC_OVERFLOW			equ	7
RC_NO_START_TAG			equ	8
RC_TAG_MISMATCH			equ	9
RC_INVALID_TAG			equ	10
;[c]
;[c]Various constants
;[c]
TAB_SIZE		equ	8
DEFAULT_CHUNK_SIZE	equ	4096
HEADER_SIZE		equ	16	; reserved space for chunks
;[cf]
;[of]:C Stubs
;[of]:initialize
;[c]Initialize the library
;[c]
;[c]	void ax_initialize(malloc, free)
;[c]
;[c]ARGUMENTS
;[c]	malloc
;[c]		the memory alllocation function
;[c]	free
;[c]		the free memory function
;[c]
;[c]
_initialize:	push	ebx
		mov	eax,[esp+8+0]
		mov	ebx,[esp+8+4]
		call	initialize
		pop	ebx
		ret
;[cf]
;[of]:initializeParser
;[c]Initialize the parse context
;[c]
;[c]	errorCode = ax_initializeParser(context, chunkSize)
;[c]
;[c]ARGUMENTS
;[c]	context
;[c]		the parse context.
;[c]	chunkSize
;[c]		the default size of chunk
;[c]
;[c]RETURN VALUE
;[c]	The error code
;[c]
_initializeParser:
		push	ebp
		mov	ebp,[esp+8+0]
		mov	ecx,[esp+8+4]
		call	initializeParser
		pop	ebp
		ret
;[cf]
;[of]:releaseParser
;[c]Release the parse context
;[c]
;[c]	ax_releaseParser(context)
;[c]
;[c]ARGUMENTS
;[c]	context
;[c]		the parse context. The memory base and memory limit must be 
;[c]		initialized.
;[c]
_releaseParser:	push	ebp
		mov	ebp,[esp+8+0]
		call	releaseParser
		pop	ebp
		ret
;[cf]
;[of]:parse
;[c]Parse an XML string
;[c]
;[c]	Element* ax_parse(context, source, type)
;[c]
;[c]ARGUMENTS
;[c]	context
;[c]		the parse context. The object must have been initialized.
;[c]	source
;[c]		the xml to parse
;[c]	type
;[c]		the expected type of element to parse
;[c]
;[c]RETURN VALUE
;[c]	The created element or null if an error occured.
;[c]
_parse:		push	ebx
		push	esi
		push	edi
		push	ebp

		mov	ebp,[esp+20+0]
		mov	esi,[esp+20+4]
		mov	edx,[esp+20+8]
		call	parse

		mov	[ebp+ParseContext.current],esi
		mov	[ebp+ParseContext.errorCode],eax
		jz	.ok
		call	computeLineColumn
		xor	eax,eax
		jmp	.error

.ok:		mov	eax,[ebp+ParseContext.root]
.error:		pop	ebp
		pop	edi
		pop	esi
		pop	ebx
		ret
;[c]
;[of]:computeLineColumn
;[c]Read the text file again nd setup line and column
;[c]
;[c]	It is faster to count lines only when an error occurs. So the parsing
;[c]	routine does not compute lines and if an error occurs, the file is
;[c]	read again just to find the line and column of the last read byte.
;[c]
computeLineColumn:

		mov	edi,esi
		mov	esi,[ebp+ParseContext.source]
		mov	ecx,1	; line number
		mov	edx,esi	; beginning of line
		cmp	esi,edi
		jz	.continue

.loop1:		mov	al,[esi]
		add	esi,1
		cmp	al,13
		jz	.CR
		cmp	al,10
		jz	.LF
.next:		cmp	esi,edi
		jnz	.loop1
		jmp	.continue
				
.CR:		cmp	esi,edi
		jz	.LF
		cmp	byte [esi],10
		jnz	.LF
		add	esi,1
.LF:		add	ecx,1
		mov	edx,esi
		jmp	.next

.continue:	mov	[ebp+ParseContext.line],ecx

		;
		; Compute column by expanding tabs
		;
		
		mov	esi,edx
		xor	ecx,ecx		; column
		mov	edx,TAB_SIZE	; next tab
		cmp	esi,edi
		jz	.continue2
.loop2:		mov	al,[esi]
		add	esi,1
		cmp	al,9
		jnz	.notTab
		mov	ecx,edx
		add	edx,TAB_SIZE
		cmp	esi,edi
		jnz	.loop2
		jmp	.continue2
.notTab:	add	ecx,1
		cmp	esi,edi
		jnz	.loop2

.continue2:	inc	ecx
		mov	[ebp+ParseContext.column],ecx

		ret
;[cf]
;[cf]
;[c]
;[of]:initializeClassParser
;[c]Initialize the class parser
;[c]
;[c]	errorCode = ax_initializeClassParser(context)
;[c]
;[c]ARGUMENTS
;[c]	context
;[c]		the class parser.
;[c]
;[c]RETURN VALUE
;[c]	The error code
;[c]
_initializeClassParser:
		push	ebp
		mov	ebp,[esp+8+0]
		call	initializeClassParser
		pop	ebp
		ret
;[cf]
;[of]:releaseClassParser
;[c]Release the class parser
;[c]
;[c]	ax_releaseClassParser(context)
;[c]
;[c]ARGUMENTS
;[c]	context
;[c]		the class parser.
;[c]
_releaseClassParser:
		push	ebp
		mov	ebp,[esp+8+0]
		call	releaseClassParser
		pop	ebp
		ret
;[cf]
;[of]:classFromElement
;[c]Create a class from an element
;[c]
;[c]	ElementClass* ax_classFromElement(element, context)
;[c]
;[c]ARGUMENTS
;[c]	element
;[c]		the xml element describing the class
;[c]	context
;[c]		the class context.
;[c]
;[c]RETURN VALUE
;[c]	The created class or null if an error occured.
;[c]
_classFromElement:
		push	esi
		push	edi
		push	ebp

		mov	esi,[esp+16]
		mov	ebp,[esp+16+4]
		call	classFromElement
		mov	[ebp+CreateClassContext.errorCode],eax
		jnz	.failed
		mov	eax,edi

.error:		pop	ebp
		pop	edi
		pop	esi
		ret

.failed:	xor	eax,eax
		jmp	.error
;[c]
;[cf]
;[of]:classFromString
;[c]Create a class from a string
;[c]
;[c]	ElementClass* ax_classFromString(string, context)
;[c]
;[c]ARGUMENTS
;[c]	string
;[c]		the xml element describing the class
;[c]	context
;[c]		the class context.
;[c]
;[c]RETURN VALUE
;[c]	The created class or null if an error occured.
;[c]
_classFromString:
		push	esi
		push	edi
		push	ebp
		
		mov	esi,[esp+16+0]
		mov	ebp,[esp+16+4]
		call	classFromString
		mov	[ebp+CreateClassContext.errorCode],eax
		jnz	.failed
		mov	eax,edi
		
.error:		pop	ebp
		pop	edi
		pop	esi
		ret

.failed:	xor	eax,eax
		jmp	.error
;[cf]
;[cf]
;[of]:Functions
;[of]:initialize
;[c]Initialize the library
;[c]
;[c]ARGUMENTS
;[c]	eax
;[c]		the allocator function
;[c]	ebx
;[c]		the free memory function
;[c]
initialize:
	mov	[mallocFunction],eax
	mov	[freeFunction],ebx
	ret
;[cf]
;[c]
;[of]:initializeParser
;[c]Initialize the parse context object
;[c]
;[c]ARGUMENTS
;[c]	ebp
;[c]		the parse context
;[c]	ecx	
;[c]		chunk size
;[c]
;[c]RETURN VALUES
;[c]	eax+zf
;[c]		error code
;[c]
initializeParser:
		xor	eax,eax
		mov	[ebp+ParseContext.chunkSize],ecx
		mov	[ebp+ParseContext.chunks],eax
		mov	eax,ecx
		jmp	newChunk
;[cf]
;[of]:releaseParser
;[c]Release the parse context object
;[c]
;[c]ARGUMENTS
;[c]	ebp
;[c]		the parse context
;[c]
releaseParser:	jmp	releaseChunks
;[cf]
;[of]:parse
;[c]parse
;[c]
;[c]ARGUMENTS
;[c]	edx
;[c]		the type of the root element
;[c]	esi
;[c]		source
;[c]	ebp
;[c]		Parse context
;[c]
;[c]RETURN VALUES
;[c]	eax+zf
;[c]		error code
;[c]	esi
;[c]		Address of the first byte after the element.
;[c]	edi
;[c]		???
;[c]
parse:		mov	[ebp+ParseContext.source],esi

		; Allocate memory for root element		
		mov	edi,[ebp+ParseContext.base]
		mov	eax,[ebp+ParseContext.limit]
		mov	ebx,edi
		add	ebx,[edx+ElementClass.size]
		cmp	ebx,eax
		jae	parseChunkFull
		mov	[ebp+ParseContext.base],ebx
		mov	[ebp+ParseContext.root],edi

resumeParseChunkFull:

		; Clear higher bits of eax
		; All the functions assume that the high bits of eax
		; are always zero.
		xor	eax,eax

		call	readProlog

		cmp	al,'<'
		jnz	notStartTag

;[c]
;[c]Sub-functions
;[c]
;[of]:	Initialize Element
		xor	al,al
		mov	[edi+Element.type],edx
		mov	[edi+Element.firstChild],eax
		mov	[edi+Element.lastChild],eax
		mov	[edi+Element.nextSibling],eax

		mov	ecx,[edx+ElementClass.propertyCount]
		lea	ebx,[edi+Element.sizeof]

		or	ecx,ecx
		jz	.m0

		shr	ecx,1
		jnc	.m2
		mov	[ebx+Attribute.begin],eax
		mov	[ebx+Attribute.limit],eax
		jz	.m0
		add	ebx,Attribute.sizeof

.m2:		shr	ecx,1
		jnc	.m4
		mov	[ebx+Attribute.begin],eax
		mov	[ebx+Attribute.limit],eax
		mov	[ebx+Attribute.sizeof+Attribute.begin],eax
		mov	[ebx+Attribute.sizeof+Attribute.limit],eax
		jz	.m0
		add	ebx,Attribute.sizeof*2

.m4:		mov	[ebx+Attribute.sizeof*0+Attribute.begin],eax
		mov	[ebx+Attribute.sizeof*0+Attribute.limit],eax
		mov	[ebx+Attribute.sizeof*1+Attribute.begin],eax
		mov	[ebx+Attribute.sizeof*1+Attribute.limit],eax
		mov	[ebx+Attribute.sizeof*2+Attribute.begin],eax
		mov	[ebx+Attribute.sizeof*2+Attribute.limit],eax
		mov	[ebx+Attribute.sizeof*3+Attribute.begin],eax
		mov	[ebx+Attribute.sizeof*3+Attribute.limit],eax
		add	ebx,Attribute.sizeof*4
		dec	ecx
		jnz	.m4
.m0:

;[cf]
;[c]		
;[of]:	Read Element Name
;[c]Read Element Name
;[c]
		mov	ebx,edi
		mov	edi,[edx+ElementClass.name]
		mov	ecx,[edx+ElementClass.nameLimit]
.loop:		cmpsb
		jnz	invalidTag
		cmp	edi,ecx
		jnz	.loop
		mov	edi,ebx
		
		mov	al,[esi]
		add	esi,1
		cmp	al,'>'
		jz	readContent
		test	byte [S + eax], 1
		jz	invalidChar
;[cf]
;[of]:	Read Attributes
;[c]Read Attributes
;[c]
;[c]Read the next attribute (skip blanks first)
;[c]
nextAttribute:
		; Skip blanks
.sb:		mov	al,[esi]
		add	esi,1
		test	byte [S + eax], 1
		jnz	.sb

;[c]
;[c]Read the next attribute (blanks already skipped, first char already read)
;[c]
readAttributes:	
		; No more attributes ?
		cmp	al, '>'
		jz	readContent
		cmp	al, '/'
		jz	endOfElement
		
;[c]
;[c]Read and Decode Attribute Name
;[c]
readAttribute:	mov	ebx,[edx+ElementClass.attributes]
		mov	ecx,[ebx+eax*4]		; first char already loaded
		test	ecx,1
		jnz	.j2
.l2:		mov	al,[esi]
		mov	ecx,[ecx+eax*4]
		add	esi,1
		test	ecx,1
		jz	.l2
		test	ecx,2
		jnz	terminalAttribute
		
.j2:		dec	ecx
		jz	invalidChar
.l21:		mov	al,[esi]
		add	esi,1
		add	ecx,1
		cmp	[ecx-1],al
		jz	.l21
		test	byte [ecx-1],255
		jnz	invalidChar

commonAttribute:
		mov	ebx,edi
		add	ebx,[ecx]
		
;[c]
;[c]Skip Blanks and Equal Sign
;[c]		
		test	byte [S + eax], 1
		jz	.notBlank2
.skipBlank2:	mov	al,[esi]
		add	esi,1
		test	byte [S + eax], 1
		jnz	.skipBlank2
.notBlank2:
		cmp	al,'='
		jnz	invalidChar

.skipBlank3:	mov	al,[esi]
		add	esi,1
		test	byte [S + eax], 1
		jnz	.skipBlank3
.notBlank3:

		cmp	al,$27	; '
		jz	.value
		cmp	al,'"'
		jnz	invalidChar
.value:		mov	cl,al

;[c]
;[c]Read and Store Value
;[c]	
		cmp	[ebx+Attribute.begin],0
		jnz	attributeDefined
		mov	[ebx+Attribute.begin],esi
		
.l3:		mov	al,[esi]
		add	esi,1
		or	al,al
		jz	unexpectedEnd
		cmp	al,'&'
		jz	attributeReference
		cmp	al,cl
		jnz	.l3
	
		lea	ecx,[esi-1]
		mov	[ebx+Attribute.limit],ecx
;[c]
endOfAttribute:
		; Read next char
		mov	al,[esi]
		add	esi,1
		
		; Is it a blank ?
		test	byte [S + eax], 1
		jnz	nextAttribute
		
		; Is it a closing tag ?
		cmp	al, '>'
		jz	readContent
		
		; Is it not the end of the element ?
		cmp	al, '/'
		jnz	readAttribute

endOfElement:	mov	al,[esi]
		add	esi,1
		cmp	al,'>'
		jnz	invalidChar
		xor	eax,eax
		ret
		
;[c]
terminalAttribute:
		and	ecx,not 2
		jmp	commonAttribute
;[cf]
;[of]:	Read Content
;[c]Read Content
;[c]
readContent:

.loop:		mov	al,[esi]
		or	al,al
		jz	unexpectedEnd
		add	esi,1
		cmp	al,'<'
		jnz	.loop

		mov	al,[esi]
		cmp	al, '/'
		jnz	referenceOrComment

;[of]:Read Close Tag
readCloseTag:
		add	esi,1

		mov	edi,[edx+ElementClass.name]
		mov	ecx,[edx+ElementClass.nameLimit]
.loop:		cmpsb
		jnz	tagMismatch
		cmp	edi,ecx
		jnz	.loop
		
		mov	al,[esi]
		add	esi,1
		cmp	al, '>'
		jnz	.skipBlanks4
		xor	eax,eax
		ret
		
.skipBlanks4:	test	byte [S + eax], 1
		jz	invalidChar
.skipBlank4:	mov	al,[esi]
		add	esi,1
		test	byte [S + eax], 1
		jnz	.skipBlank4
		cmp	al, '>'
		jnz	invalidChar
		
		xor	eax,eax		; element successfully read
		ret

;[cf]

referenceOrComment:
		cmp	al,'!'
		jz	skipComment

;[of]:Read Child Element
;[c]Read a child element (may be a text-only element)
;[c]
readChild:	add	esi,1
;[c]
;[c]Decode Element Name
;[c]
		mov	ebx,[edx+ElementClass.elements]
		mov	ecx,[ebx+eax*4]		; first char already loaded
		test	ecx,1
		jnz	.skip
.loop:		mov	al,[esi]
		mov	ecx,[ecx+eax*4]
		add	esi,1
		test	ecx,1
		jz	.loop
		
.skip:		test	ecx,2
		jnz	terminalElement
		dec	ecx
		jz	invalidChar
.loop2:		mov	al,[esi]
		add	ecx,1
		add	esi,1
		cmp	[ecx-1],al
		jz	.loop2
		test	byte [ecx-1],255
		jnz	invalidChar

		mov	bl,[ecx]

;[of]:Skip Blanks
		test	byte [S + eax], 1
		jz	.eNotBlank1
.eSkipBlank1:	mov	al,[esi]
		add	esi,1
		test	byte [S + eax], 1
		jnz	.eSkipBlank1
.eNotBlank1:
;[cf]

		add	ecx,3+1
		and	ecx,$FFFFFFFC
		test	bl,1
		jz	readChildElement

;[c]
;[of]:Read Text
;[c]Read a Text-Only Element
;[c]	
;[c]	* esi is the source
;[c]	* edi is the element
;[c]
;[c]	* ecx is the text element descriptor
;[c]	* al is the last non blank character in the tag (must be '>')
;[c]
readText:
		mov	ebx,edi
		add	ebx,[ecx+TextElementClass.offset]
		
		cmp	al,'/'
		jz	emptyText
		cmp	al,'>'
		jnz	invalidChar

		mov	[ebx+Attribute.begin],esi

		mov	al,[esi]
		add	esi,1
		cmp	al,'<'
		jz	emptyOrCDATA
		or	al,al
		jz	unexpectedEnd
		cmp	al,'&'
		jz	textReference

.loop:		mov	al,[esi]
		add	esi,1
		or	al,al
		jz	unexpectedEnd
		cmp	al,'&'
		jz	textReference
		cmp	al,'<'
		jnz	.loop

		mov	al,[esi]
		cmp	al,'!'
		jz	textCommentOrCDATA

endOfEmptyText:	dec	esi
		mov	[ebx+Attribute.limit],esi
		inc	esi

endOfText:	add	esi,1
		cmp	al, '/'
		jnz	invalidChar

		mov	ebx,edi
		mov	edi,[ecx+TextElementClass.begin]
		mov	ecx,[ecx+TextElementClass.limit]
.loop:		cmpsb
		jnz	tagMismatch
		cmp	edi,ecx
		jnz	.loop
		mov	edi,ebx
		
.eSkipBlank2:	mov	al,[esi]
		add	esi,1
		test	byte [S + eax], 1
		jnz	.eSkipBlank2
		
		cmp	al,'>'
		jz	readContent
		jmp	invalidChar
;[c]
emptyOrCDATA:	mov	al,[esi]
		cmp	al,'!'
		jnz	endOfEmptyText
		cmp	byte [esi+1],'['
		jnz	invalidChar
		cmp	byte [esi+2],'C'
		jnz	invalidChar
		cmp	byte [esi+3],'D'
		jnz	invalidChar
		cmp	byte [esi+4],'A'
		jnz	invalidChar
		cmp	byte [esi+5],'T'
		jnz	invalidChar
		cmp	byte [esi+6],'A'
		jnz	invalidChar
		cmp	byte [esi+7],'['
		jnz	invalidChar
		add	esi,8
		mov	[ebx+Attribute.begin],esi

.loop:		mov	al,[esi]
		add	esi,1
		or	al,al
		jz	unexpectedEnd
		cmp	al,']'
		jnz	.loop
		mov	al,[esi]
		cmp	al,']'
		jnz	.loop
		mov	al,[esi+1]
		cmp	al,'>'
		jnz	.loop

		mov	al,[esi+2]
		cmp	al,'<'
		jnz	.mixed

		sub	esi,1
		mov	[ebx+Attribute.limit],esi
		add	esi,4
		mov	al,[esi]
		jmp	endOfText
;[c]
;[c]The text starts with a CDATA section but contain something else,
;[c]let 's read it again with the slow method
;[c]
.mixed:		mov	esi,[ebx+Attribute.begin]
		sub	esi,9		; size("<![CDATA[")
		mov	[ebx+Attribute.begin],esi
		jmp	textRead
;[c]
emptyText:	mov	al,[esi]
		add	esi,1
		cmp	al,'>'
		jnz	invalidChar
		mov	[ebx+Attribute.begin],esi	; any non null ptr is ok
		mov	[ebx+Attribute.limit],esi
		jmp	readContent
;[cf]
;[of]:Read a Child Element
;[c]Read a Child Element
;[c]
;[c]	* esi source
;[c]	* edi parent element
;[c]	* ecx the element description in parent
;[c]	* edx type of parent
;[c]

readChildElement:
		
		push	edx
		mov	edx,[ecx]
		
		; Allocate memory
		mov	ecx,[ebp+ParseContext.base]
		mov	ebx,[edx+ElementClass.size]
		add	ebx,ecx
		cmp	ebx,[ebp+ParseContext.limit]
		jae	.chunkFull
		mov	[ebp+ParseContext.base],ebx
.resume:
		; Append element
		cmp	dword [edi+Element.firstChild],0
		jnz	.rce1
		mov	[edi+Element.firstChild],ecx
.rce1:		mov	ebx,[edi+Element.lastChild]
		or	ebx,ebx
		jz	.rce2
		mov	[ebx+Element.nextSibling],ecx
.rce2:		mov	[edi+Element.lastChild],ecx

		push	edi
		mov	edi,ecx
;[of]:	Initialize Element
		push	eax
		xor	al,al
		mov	[edi+Element.type],edx
		mov	[edi+Element.firstChild],eax
		mov	[edi+Element.lastChild],eax
		mov	[edi+Element.nextSibling],eax

		mov	ecx,[edx+ElementClass.propertyCount]
		lea	ebx,[edi+Element.sizeof]

		or	ecx,ecx
		jz	.xm0

		shr	ecx,1
		jnc	.xm2
		mov	[ebx+Attribute.begin],eax
		mov	[ebx+Attribute.limit],eax
		jz	.xm0
		add	ebx,Attribute.sizeof

.xm2:		shr	ecx,1
		jnc	.xm4
		mov	[ebx+Attribute.begin],eax
		mov	[ebx+Attribute.limit],eax
		mov	[ebx+Attribute.sizeof+Attribute.begin],eax
		mov	[ebx+Attribute.sizeof+Attribute.limit],eax
		jz	.xm0
		add	ebx,Attribute.sizeof*2

.xm4:		mov	[ebx+Attribute.sizeof*0+Attribute.begin],eax
		mov	[ebx+Attribute.sizeof*0+Attribute.limit],eax
		mov	[ebx+Attribute.sizeof*1+Attribute.begin],eax
		mov	[ebx+Attribute.sizeof*1+Attribute.limit],eax
		mov	[ebx+Attribute.sizeof*2+Attribute.begin],eax
		mov	[ebx+Attribute.sizeof*2+Attribute.limit],eax
		mov	[ebx+Attribute.sizeof*3+Attribute.begin],eax
		mov	[ebx+Attribute.sizeof*3+Attribute.limit],eax
		add	ebx,Attribute.sizeof*4
		dec	ecx
		jnz	.xm4

.xm0:		pop	eax

;[cf]
		call	readAttributes
		pop	edi
		pop	edx
		jz	readContent
		ret
;[c]
.chunkFull:	push	eax
		push	edx
		push	edi
		mov	eax,[edx+ElementClass.size]
		call	extendMemory
		mov	ecx,edi
		pop	edi
		pop	edx
		pop	eax
		jz	.resume
		pop	edx
		jmp	parseOutOfMemory
;[cf]
;[c]
terminalElement:
		add	ecx,1
		mov	bl,[ecx-3]

;[of]:Skip Blanks
		test	byte [S + eax], 1
		jz	.eNotBlank1
.eSkipBlank1:	mov	al,[esi]
		add	esi,1
		test	byte [S + eax], 1
		jnz	.eSkipBlank1
.eNotBlank1:
;[cf]

		test	bl,1
		jz	readChildElement
		jmp	readText		
;[cf]
;[of]:Skip comment
skipComment:	add	esi,1
		mov	al,[esi]
		add	esi,1
		cmp	al,'-'
		jnz	invalidChar
		mov	al,[esi]
		add	esi,1
		cmp	al,'-'
		jnz	invalidChar
.loop:		mov	al,[esi]
		add	esi,1
		or	al,al
		jz	unexpectedEnd
		cmp	al,'-'
		jnz	.loop
		mov	al,[esi]
		add	esi,1
		cmp	al,'-'
		jnz	.loop
		mov	al,[esi]
		add	esi,1
		cmp	al,'>'
		jz	readContent
		jmp	invalidChar
;[cf]
;[cf]
;[c]
;[of]:	attributeReference
attributeReference:
		push	ecx
		push	edx
		push	edi
;[c]
;[c]Count chars
;[c]
		mov	ah,cl
		lea	ecx,[esi-1]
.l1:		mov	al,[esi]
		add	esi,1
		or	al,al
		jz	.unexpectedEnd
		cmp	al,ah
		jnz	.l1
		mov	edx,esi
		mov	esi,[ebx+Attribute.begin]
		sub	ecx,esi	; length of first part
		sub	edx,esi	; maximum length
;[c]		
;[c]Pre-reserve sufficient space
;[c]		
		mov	edi,[ebp+ParseContext.base]
		mov	[ebx+Attribute.begin],edi
		add	edx,edi
		cmp	edx,[ebp+ParseContext.limit]
		jae	.chunkFull
		mov	[ebp+ParseContext.base],edx
.resume:
;[c]
;[c]Copy first part
;[c]
		rep 
		movsb
		add	esi,1
		
		mov	cl,ah	; restore sting delimiter
		xor	ah,ah	; restore ah to 0
;[c]
;[c]Copy remaining chars
;[c]
;[c]	Note: it is no longer to check for unexoected end in this loop since
;[c]	we have already scanned the value while counting chars.
;[c]		
.decode:	call	decodeReference
.loop:		stosb
		mov	al,[esi]
		add	esi,1
		cmp	al,'&'
		jz	.decode
		cmp	al,cl
		jnz	.loop

		mov	[ebx+Attribute.limit],edi
;[c]		
;[c]Free remaining space (align on the next dword)
;[c]
		add	edi,3
		and	edi,not 3
		mov	[ebp+ParseContext.base],edi
		pop	edi
		pop	edx
		pop	ecx

		; Terminate this attribute and go to the next one
		jmp	endOfAttribute
;[c]
.chunkFull:	push	ecx
		push	edx
		mov	eax,edx
		sub	eax,edi
		call	extendMemory
		pop	edx
		pop	ecx
		jz	.resume

		pop	edi
		pop	edx
		pop	ecx
		jmp	parseOutOfMemory
;[c]
.unexpectedEnd:	pop	edi
		pop	edx
		pop	ecx
		jmp	unexpectedEnd
;[cf]
;[of]:	textReference
textCommentOrCDATA:
		sub	esi,1

textReference:	sub	esi,1

textRead:	push	ecx
		push	edx
		push	edi
;[c]
;[c]Count chars
;[c]
		mov	ecx,esi
		mov	esi,[ebx+Attribute.begin]
		sub	ecx,esi	; length of first part
;[c]		
;[c]Pre-reserve sufficient space
;[c]		
		mov	edi,[ebp+ParseContext.base]
		mov	[ebx+Attribute.begin],edi
		lea	edx,[edi+ecx]
		cmp	edx,[ebp+ParseContext.limit]
		jae	.chunkFull
.resume:
;[c]
;[c]Copy first part
;[c]
		rep movsb
;[c]
;[c]Copy remaining chars
;[c]
;[c]	Note: it is no longer to check for unexpected end in this loop since
;[c]	we have already scanned the value while counting chars.
;[c]		
		mov	ecx,[ebp+ParseContext.limit]
		sub	ecx,edi
		jmp	.decodeContent

.decode:	call	decodeReference
.loop:		stosb
		sub	ecx,1
		jz	.realloc1
.decodeContent:	mov	al,[esi]
		add	esi,1
		cmp	al,'&'
		jz	.decode
		cmp	al,'<'
		jnz	.loop
		mov	al,[esi]
		cmp	al,'!'
		jz	.CDATAOrComment

		mov	[ebx+Attribute.limit],edi
;[c]		
;[c]Free remaining space (align on the next dword)
;[c]
		add	edi,3
		and	edi,not 3
		mov	[ebp+ParseContext.base],edi
		pop	edi
		pop	edx
		pop	ecx

		; Terminate this attribute and go to the next one
		mov	al,[esi]
		jmp	endOfText
;[c]
.chunkFull:	push	ecx
		lea	eax,[ecx+1] ; at least one extra char
		call	extendMemory
		pop	ecx
		jnz	.outOfMemory
		jmp	.resume
;[c]
.realloc1:	call	.realloc
		jmp	.decodeContent
;[c]
.realloc:	mov	eax,edi
		sub	eax,[ebx+Attribute.begin]
		add	eax,[ebp+MemoryContext.chunkSize]
		add	eax,3
		and	eax,not 3
		call	extendMemory
		jnz	.outOfMemory1

		push	esi
		mov	ecx,esi
		mov	esi,[ebx+Attribute.begin]
		mov	edi,[ebp+ParseContext.base]
		mov	[ebx+Attribute.begin],edi
		sub	ecx,esi
		rep movsb
		add	esi,1
		mov	ecx,[ebp+ParseContext.limit]
		sub	ecx,edi
		pop	esi		
		ret
;[c]
.outOfMemory1:	add	esp,4
.outOfMemory:	pop	edi
		pop	edx
		pop	ecx
		jmp	parseOutOfMemory
;[c]
.unexpectedEnd:	pop	edi
		pop	edx
		pop	ecx
		jmp	unexpectedEnd
;[c]
.invalidChar:	pop	edi
		pop	edx
		pop	ecx
		jmp	invalidChar
;[c]
.CDATAOrComment:
		add	esi,1
		mov	al,[esi]
		add	esi,1
		cmp	al,'-'
		jnz	.CDATA

;[of]:Skip Comments
		mov	al,[esi]
		add	esi,1
		cmp	al,'-'
		jnz	.invalidChar
.sc:		mov	al,[esi]
		add	esi,1
		or	al,al
		jz	.unexpectedEnd
		cmp	al,'-'
		jnz	.sc
		mov	al,[esi]
		add	esi,1
		cmp	al,'-'
		jnz	.sc
		mov	al,[esi]
		add	esi,1
		cmp	al,'>'
		jz	.decodeContent
		jmp	.invalidChar
;[cf]
;[of]:Read CDATA Section
.CDATA:		cmp	al,'['
		jnz	.invalidChar
		cmp	byte [esi],'C'
		jnz	.invalidChar
		cmp	byte [esi+1],'D'
		jnz	.invalidChar
		cmp	byte [esi+2],'A'
		jnz	.invalidChar
		cmp	byte [esi+3],'T'
		jnz	.invalidChar
		cmp	byte [esi+4],'A'
		jnz	.invalidChar
		cmp	byte [esi+5],'['
		jnz	.invalidChar
		add	esi,6

		jmp	.cd2
.cd1:		stosb
		sub	ecx,1
		jz	.cd3		; chunk is full - realloc
.cd2:		mov	al,[esi]
		add	esi,1
		or	al,al
		jz	.unexpectedEnd
		cmp	al,']'
		jnz	.cd1
		mov	al,[esi]
		cmp	al,']'
		jnz	.cd1
		mov	al,[esi+1]
		cmp	al,'>'
		jnz	.cd1
		add	esi,2
		jmp	.decodeContent

.cd3:		call	.realloc
		jmp	.cd2
;[cf]
;[cf]
;[of]:	decodeReference
decodeReference:
		mov	al,[esi]
		cmp	al,'#'
		jz	.num
		cmp	al,'a'
		jz	.a
		cmp	al,'l'
		jz	.l
		cmp	al,'g'
		jz	.g
		cmp	al,'q'
		jnz	.invalidEntityReference

		; &quot;
.q:		cmp	byte [esi+1],'u'
		jnz	.invalidEntityReference
		cmp	byte [esi+2],'o'
		jnz	.invalidEntityReference
		cmp	byte [esi+3],'t'
		jnz	.invalidEntityReference
		cmp	byte [esi+4],';'
		jnz	.invalidEntityReference
		add	esi,5
		mov	al,'"'
		ret

.num:		add	esi,1
		xor	edx,edx
		mov	al,[esi]
		add	esi,1
		cmp	al,'x'
		jz	.hexa
		test	byte [digit+eax],1
		jz	.invalidChar
.nextDigit:	sub	al,'0'
		imul	edx,10
		jc	.overflow
		add	edx,eax		
		jc	.overflow
		mov	al,[esi]
		add	esi,1
		test	byte [digit+eax],1
		jnz	.nextDigit
		cmp	al,';'
		jnz	.invalidEntityReference
		mov	eax,edx
		test	edx,$FFFFFF00
		jnz	.invalidChar
		ret

.hexa:		mov	al,[esi]
		add	esi,1
		test	byte [digit+eax],2
		jz	.invalidChar
.nextXDigit:	shl	edx,4
		jc	.overflow
		or	dl,[xdigit+eax]
		mov	al,[esi]
		add	esi,1
		test	byte [digit+eax],2
		jnz	.nextXDigit
		cmp	al,';'
		jnz	.invalidEntityReference
		mov	eax,edx
		test	edx,$FFFFFF00
		jnz	.invalidChar
		ret
		
.a:		mov	al,[esi+1]
		cmp	al,'p'
		jz	.ap
		cmp	al,'m'
		jnz	.invalidEntityReference

		; &amp;
.am:		cmp	byte [esi+2],'p'
		jnz	.invalidEntityReference
		cmp	byte [esi+3],';'
		jnz	.invalidEntityReference
		add	esi,4
		mov	al,'&'
		ret

		; &apos;
.ap:		cmp	byte [esi+2],'o'
		jnz	.invalidEntityReference
		cmp	byte [esi+3],'s'
		jnz	.invalidEntityReference
		cmp	byte [esi+4],';'
		jnz	.invalidEntityReference
		add	esi,5
		mov	al,$27
		ret

		; &lt;
.l:		cmp	byte [esi+1],'t'
		jnz	.invalidEntityReference
		cmp	byte [esi+2],';'
		jnz	.invalidEntityReference
		add	esi,3
		mov	al,'<'
		ret

		; &gt;
.g:		cmp	byte [esi+1],'t'
		jnz	.invalidEntityReference
		cmp	byte [esi+2],';'
		jnz	.invalidEntityReference
		add	esi,3
		mov	al,'>'
		ret
;[c]
.invalidChar:	add	esp,4
		pop	edi
		pop	edx
		pop	ecx
		jmp	invalidChar
;[c]
.overflow:	add	esp,4
		pop	edi
		pop	edx
		pop	ecx
		jmp	overflow
;[c]
.invalidEntityReference:
		add	esp,4
		pop	edi
		pop	edx
		pop	ecx
		jmp	invalidEntityReference
;[cf]
;[c]
;[of]:	readProlog
;[c]Read the prolog
;[c]
readProlog:
		call	readXMLDecl

		; Skip blanks
.sb:		mov	al,[esi]
		add	esi,1
		test	byte [S + eax], 1
		jnz	.sb
		
		; Skip comment
		cmp	al,'<'
		jnz	.done
		cmp	byte [esi],'!'
		jnz	.done
		cmp	byte [esi+1],'-'
		jnz	.done
		cmp	byte [esi+2],'-'
		jnz	.done
		add	esi,3
.sc:		mov	al,[esi]
		add	esi,1
		or	al,al
		jz	unexpectedEnd
		cmp	al,'-'
		jnz	.sc
		mov	al,[esi]
		add	esi,1
		cmp	al,'-'
		jnz	.sc
		mov	al,[esi]
		add	esi,1
		cmp	al,'>'
		jz	.sb
		
		add	esp,4
		jmp	invalidChar

.done:		ret
;[c]
;[c]SUB-FUNCTIONS
;[of]:	readXMLDecl
;[c]
readXMLDecl:	
		cmp	byte [esi],'<'
		jnz	.done
		cmp	byte [esi+1],'?'
		jnz	.done
		cmp	byte [esi+2],'x'
		jnz	.done
		cmp	byte [esi+3],'m'
		jnz	.done
		cmp	byte [esi+4],'l'
		add	esi,5

.loop:		mov	al,[esi]
		add	esi,1
		cmp	al,'?'
		jnz	.loop
		mov	al,[esi]
		cmp	al,'>'
		jnz	.loop
		add	esi,1
		;jmp	.done
		;add	esp,8
		;jmp	invalidChar

.done:		ret
;[cf]
;[cf]
;[c]
;[c]Recoverable Errors...	
;[c]
;[of]:	parseChunkFull
parseChunkFull:	push	ecx
		push	edx
		mov	eax,[edx+ElementClass.size]
		call	extendMemory
		pop	edx
		pop	ecx
		jz	resumeParseChunkFull
		ret
;[cf]
;[c]
;[c]Errors...
;[c]
;[of]:	invalidEntityReference
invalidEntityReference:
		mov	eax,RC_INVALID_ENTITY_REFERENCE
		or	eax,eax
		ret
;[cf]
;[of]:	unexpectedEnd
unexpectedEnd:	mov	eax,RC_UNEXPECTED_END
		or	eax,eax
		ret
;[cf]
;[of]:	invalidChar
invalidChar:	mov	eax,RC_INVALID_CHAR
		or	eax,eax
		ret
;[cf]
;[of]:	overflow
overflow:	mov	eax,RC_OVERFLOW
		or	eax,eax
		ret
;[cf]
;[of]:	notStartTag
notStartTag:	mov	eax,RC_NO_START_TAG
		or	eax,eax
		ret
;[cf]
;[of]:	tagMismatch
tagMismatch:	mov	eax,RC_TAG_MISMATCH
		or	eax,eax
		ret
;[cf]
;[of]:	invalidTag
invalidTag:	mov	eax,RC_INVALID_TAG
		or	eax,eax
		ret
;[cf]
;[of]:	parseOutOfMemory
parseOutOfMemory:
		mov	eax,RC_MEMORY
		or	eax,eax
		ret
;[cf]
;[of]:	attributeDefined
attributeDefined:
		mov	eax,RC_ATTR_DEFINED
		or	eax,eax
		ret
;[cf]
;[cf]
;[c]
;[of]:initializeClassParser
;[c]Initialize the class parser
;[c]
;[c]ARGUMENTS
;[c]	ebp
;[c]		the parse context
;[c]
;[c]RETURN VALUES
;[c]	eax+zf
;[c]		error code
;[c]
initializeClassParser:
		mov	eax,DEFAULT_CHUNK_SIZE
		xor	ecx,ecx
		mov	[ebp+ParseContext.chunkSize],eax
		mov	[ebp+ParseContext.chunks],ecx
		jmp	newChunk
;[cf]
;[of]:releaseClassParser
;[c]Release the class parser
;[c]
;[c]ARGUMENTS
;[c]	ebp
;[c]		the class parser
;[c]
releaseClassParser:
		jmp	releaseChunks
;[cf]
;[of]:classFromString
;[c]Create a class from a string
;[c]
;[c]ARGUMENTS
;[c]	ebp
;[c]		the xml element describing the class
;[c]	ebp
;[c]		the class context.
;[c]
;[c]RETURN VALUE
;[c]	eax+zf
;[c]		error code
;[c]	edi
;[c]		the class
;[c]
classFromString:
		push	esi
		push	ebp
		mov	edi,ebp
		
		sub	esp,ParseContext.sizeof
		mov	ebp,esp
		mov	ecx,16384
		call	initializeParser
		jnz	.error

		mov	edx,DocumentClass
		push	edi
		call	parse
		pop	edi
		jnz	.error2

		mov	esi,[ebp+ParseContext.root]
		mov	ebp,edi
		call	classFromElement
		mov	ebp,esp
	
.error2:	push	eax
		call	releaseParser
		pop	eax

.error:		add	esp,ParseContext.sizeof
		or	eax,eax
		pop	ebp
		pop	esi
		ret
;[cf]
;[of]:classFromElement
;[c]Create a class from an element
;[c]
;[c]ARGUMENTS
;[c]	ebp
;[c]		context
;[c]	esi
;[c]		root element
;[c]
;[c]RETURN VALUES
;[c]	eax+zf
;[c]		error code
;[c]	edi
;[c]		the class
;[c]
classFromElement:
		push	ebx
		mov	ebx,ebp
		call	createOneClass
		pop	ebx
		ret
;[c]
;[c]SUB-FUNCTIONS
;[of]:	createOneClass
;[c]Create one class
;[c]
;[c]ARGUMENTS
;[c]	ebx
;[c]		context
;[c]	esi
;[c]		root element
;[c]
;[c]RETURN VALUES
;[c]	eax+zf
;[c]		error code
;[c]	edi
;[c]		the class
;[c]
		struct	TreeBuilder
.context	rd	1	; the create class context
.rootElement	rd	1	; the root element
.root		rd	1	; the root node
.offset		rd	1	; the attribute offset
.propertyCount	rd	1	; number of attributes/text element
.childCount	rd	1	; number of child classes
.childIndex	rd	1	; index to store next child class

.collectionClass	rd	1	; 
.nameBegin	rd	1	;
.nameLimit	rd	1	;
		ends

createOneClass:	push	ebx
		push	esi
		push	ebp
		sub	esp,TreeBuilder.sizeof
		mov	ebp,esp
		mov	[ebp+TreeBuilder.context],ebx
		mov	[ebp+TreeBuilder.rootElement],esi
		xor	eax,eax
		mov	[ebp+TreeBuilder.root],eax
		mov	[ebp+TreeBuilder.propertyCount],eax
		mov	[ebp+TreeBuilder.childIndex],eax
		mov	[ebp+TreeBuilder.offset],Element.sizeof
		call	countChildren
		mov	[ebp+TreeBuilder.childCount],eax
		
		;
		; Allocate memory for element class
		;
		mov	edx,eax
		mov	ebx,[esi+Element.sizeof+Attribute.begin]
		mov	eax,[esi+Element.sizeof+Attribute.limit]
		sub	eax,ebx
		push	eax
		add	eax,1
		lea	eax,[eax+edx*4]
		push	edx
		call	allocateElementClass
		pop	edx
		pop	ecx
		jnz	.error

		;
		; Copy and Set Name
		;
		lea	eax,[edi+ElementClass.sizeof+edx*4]
		mov	[edi+ElementClass.name],eax
		push	edi
		mov	esi,ebx
		mov	edi,eax
		rep movsb
		mov	esi,edi
		xor	eax,eax
		stosb
		pop	edi
		mov	[edi+ElementClass.nameLimit],esi

		;
		; Build the attribute tree and
		; build the attribute map from the tree
		;
		mov	esi,[ebp+TreeBuilder.rootElement]
		call	buildAttributeTree
		jnz	.error
		mov	ebx,-1
		mov	esi,[ebp+TreeBuilder.root]
		call	createAttributeMap
		jnz	.error
		mov	[edi+ElementClass.attributes],ebx

		;
		; Build the element tree
		;
		xor	eax,eax
		mov	[ebp+TreeBuilder.root],eax
		mov	esi,[ebp+TreeBuilder.rootElement]
		call	buildElementTree
		jnz	.error
		xor	ebx,ebx
		mov	esi,[ebp+TreeBuilder.root]
		call	createElementMap
		jnz	.error
		mov	[edi+ElementClass.elements],ebx
		
		;
		; Finalize:
		; - copy the number of properties
		; - setup the size of the element
		;
		mov	eax,[ebp+TreeBuilder.propertyCount]
		imul	ebx,eax,Attribute.sizeof
		add	ebx,Element.sizeof
		mov	ecx,[ebp+TreeBuilder.childCount]
		mov	[edi+ElementClass.propertyCount],eax
		mov	[edi+ElementClass.size],ebx
		mov	[edi+ElementClass.childCount],ecx

		xor	eax,eax
		
.error:		add	esp,TreeBuilder.sizeof
		or	eax,eax
		pop	ebp
		pop	esi
		pop	ebx
		ret
;[c]
;[c]SUB-FUNCTIONS
;[of]:countChildren
;[c]Count number of child classes
;[c]
;[c]ARGUMENTS
;[c]	esi
;[c]		the element
;[c]
;[c]RETURN VALUES
;[c]	eax
;[c]		the number of child classes
;[c]
countChildren:	xor	eax,eax
		mov	edx,[esi+Element.firstChild]
		jmp	.first
.loop:		cmp	[edx+Element.type],CollectionClass
		jnz	.next
		inc	eax
.next:		mov	edx,[edx+Element.nextSibling]
.first:		or	edx,edx
		jnz	.loop
		ret
;[cf]
;[of]:	copyString
copyString:	push	esi
		push	edi
		rep movsb
		xor	eax,eax
		stosb
		pop	edi
		pop	esi
		ret
;[cf]
;[of]:	allocateElementClass
;[c]Allocate memory for an element class
;[c]
;[c]ARGUMENTS
;[c]	eax
;[c]		length of name
;[c]
;[c]RETURN VALUES
;[c]	edi
;[c]		class
;[c]
allocateElementClass:
		mov	ecx,4
		add	eax,ElementClass.sizeof
		mov	edx,[ebp+TreeBuilder.context]
		jmp	allocateMemory
;[cf]
;[of]:	buildAttributeTree
;[c]Build the tree of attributes
;[c]
;[c]ARGUMENTS
;[c]	esi
;[c]		root element
;[c]
;[c]RETURN VALUES
;[c]	eax+zf
;[c]		error code
;[c]
;[c]REMARKS
;[c]	The 'root' local variable will contain the root attribute map
;[c]
buildAttributeTree:
		push	esi

		;
		; Process all attributes of the element
		;
		mov	esi,[esi+Element.firstChild]
		jmp	.first

.lp:		cmp	[esi+Element.type],AttributeClass
		jnz	.next
		xor	eax,eax
		call	addItem
		jnz	.error
.next:		mov	esi,[esi+Element.nextSibling]
.first:		or	esi,esi
		jnz	.lp
		
		xor	eax,eax
.error:		pop	esi
		ret
;[cf]
;[of]:	buildlementTree
;[c]Build the tree of elements
;[c]
;[c]ARGUMENTS
;[c]	esi
;[c]		root element
;[c]	edi
;[c]		current element class
;[c]
;[c]RETURN VALUES
;[c]	eax+zf
;[c]		error code
;[c]
;[c]REMARKS
;[c]	The 'root' local variable will contain the root attribute map
;[c]
buildElementTree:
		push	ebx
		push	esi
		push	edi
		mov	ebx,edi

		;
		; Process all attributes of the element
		;
		mov	esi,[esi+Element.firstChild]
		jmp	.first

.loop:		cmp	[esi+Element.type],TextClass
		jz	.txt
		cmp	[esi+Element.type],CollectionClass
		jz	.collection
.next:		mov	esi,[esi+Element.nextSibling]
.first:		or	esi,esi
		jnz	.loop
		
		xor	eax,eax
.error:		pop	edi
		pop	esi
		pop	ebx
		ret
;[c]
;[c]Process a <text> element
;[c]
.txt:		xor	eax,eax
		call	addItem
		jnz	.error
		jmp	.next
;[c]
;[c]Process a <collection> element
;[c]
.collection:
		; Create subclass
		push	ebx
		mov	ebx,[ebp+TreeBuilder.context]
		call	createOneClass
		pop	ebx
		jnz	.error
		
		; Save it to the list
		mov	eax,[ebp+TreeBuilder.childIndex]
		mov	[ebx+ElementClass.sizeof+eax*4],edi
		inc	eax
		mov	[ebp+TreeBuilder.childIndex],eax
		
		; Add it to the tree
		mov	eax,edi
		call	addItem
		jnz	.error

		jmp	.next
;[cf]
;[of]:	addItem
;[c]Add one item to the tree
;[c]
;[c]ARGUMENTS
;[c]	eax
;[c]		the collection class or null
;[c]	esi
;[c]		the element
;[c]	ebp
;[c]		local variables
;[c]
;[c]RETURN VALUES
;[c]	eax+zf
;[c]		error code
;[c]
;[c]REAMRKS
;[c]	It can be a text, attribute or collection item.
;[c]
addItem:	push	ebx
		push	esi
		push	edi
		push	ebp
		mov	ecx,[esi+Element.sizeof+Attribute.limit]
		mov	edx,[esi+Element.sizeof+Attribute.begin]
		mov	[ebp+TreeBuilder.collectionClass],eax
		mov	[ebp+TreeBuilder.nameLimit],ecx
		mov	[ebp+TreeBuilder.nameBegin],edx
		
		; empty name ?
		cmp	ecx,edx
		jz	.emptyName

		; esi = list
		; ebx = previous node
		mov	esi,[ebp+TreeBuilder.root]
		xor	ebx,ebx

		; read char
.next:		mov	al,[edx]
		inc	edx
		
		push	ecx
		push	edx
		call	findOrCreateNode
		pop	edx
		pop	ecx
		jnz	.error
		; esi: list
		; edi: node

		; Attach to previous node
		or	ebx,ebx
		jz	.root
		cmp	[ebx+Node.succ],0
		jnz	.root
		mov	[ebx+Node.succ],esi
.root:

		; Save the root list
		cmp	[ebp+TreeBuilder.root],0
		jnz	.hasRoot
		mov	[ebp+TreeBuilder.root],esi
.hasRoot:

		mov	ebx,edi			; the previous node is the current one
		mov	esi,[ebx+Node.succ]	; get the successor
		cmp	ecx,edx
		jnz	.next
		
		; Save the attribute offset
		; and update it
		mov	eax,[ebp+TreeBuilder.offset]
		mov	ecx,[ebp+TreeBuilder.collectionClass]
		
		cmp	dword [ebx+Node.index],-1
		jnz	.conflict
		cmp	dword [ebx+Node.collectionClass],0
		jnz	.conflict
		or	ecx,ecx
		jnz	.class
		mov	[ebx+Node.index],eax
		add	eax,Attribute.sizeof
		mov	[ebp+TreeBuilder.offset],eax
		add	[ebp+TreeBuilder.propertyCount],1
		jmp	.common
.class:		mov	[ebx+Node.collectionClass],ecx
.common:	mov	ecx,[ebp+TreeBuilder.nameBegin]
		mov	edx,[ebp+TreeBuilder.nameLimit]
		call	newDString
		mov	[ebx+Node.nameBegin],ecx
		mov	[ebx+Node.nameLimit],edx
		
		xor	eax,eax
.error:		or	eax,eax
		pop	ebp
		pop	edi
		pop	esi
		pop	ebx
		ret

.emptyName:	mov	eax,RC_EMPTY_NAME
		jmp	.error

.conflict:	mov	eax,RC_ATTR_DEFINED
		jmp	.error
;[c]
;[of]:newDString
;[c]Create a new DString by copying another one
;[c]
;[c]ARGUMENTS
;[c]	ecx
;[c]		begin
;[c]	edx
;[c]		end
;[c]
;[c]RETURN VALUES
;[c]	eax+zf
;[c]		error code
;[c]	ecx
;[c]		begin (copy)
;[c]	edx
;[c]		end (copy)
;[c]
newDString:	push	esi
		push	edi
		
		mov	esi,ecx
		mov	eax,edx
		sub	eax,esi
		mov	ecx,1
		mov	edx,[ebp+TreeBuilder.context]
		push	eax
		call	allocateMemory
		pop	ecx
		jnz	.error
		push	ecx
		push	edi
		rep 
		movsb
		pop	ecx
		pop	edx
		add	edx,ecx
		
		xor	eax,eax
.error:		pop	edi
		pop	esi
		ret
;[cf]
;[cf]
;[of]:	createElementMap
;[c]Create an element map for one node
;[c]
;[c]ARGUMENTS
;[c]	ebx
;[c]		the previous node or nil
;[c]	esi
;[c]		the node
;[c]	ebp
;[c]	
;[c]RETURN VALUE
;[c]	eax+zf
;[c]		error code
;[c]	ebx
;[c]		the map
;[c]
createElementMap:
		push	esi
		push	edi

		; Empty node ?
		or	esi,esi
		jz	.empty

		call	createMap
		
		; If the previous node contains an offset, create the
		; attribute and make any non-name char a
		; target to this attribute.
		;
		; Note: it cannot happen on the root node since
		; a name cannot be empty.
		or	ebx,ebx
		jz	.noAttr
		cmp	[ebx+Node.collectionClass],0
		jnz	.elem
		cmp	[ebx+Node.index],-1
		jz	.noAttr
.elem:
		; Case where a name is a prefix of another name:
		;   ab
		;   abc
		; ab must be added on non-chars of the abc's map
		;
		;mov	eax,8
		;mov	ecx,4
		;push	edi
		;mov	edx,[ebp+TreeBuilder.context]
		;call	allocateMemory
		;pop	edx
		;jnz	.error
		;mov	ecx,[ebx+Node.collectionClass]
		;mov	word [edi],0
		;mov	[edi+4],ecx
		
		push	ebx
		push	esi
		mov	esi,ebx
		call	createTerminalElement
		mov	ecx,ebx
		pop	esi
		pop	ebx
		jnz	.error
		
		push	edi
		push	edx
		mov	edx,edi
		lea	edi,[ecx+3] ; set bit 0 & 1 to 1
		call	mapTerminal
		pop	edx
		pop	edi
.noAttr:

		; Process each node
.lp:		cmp	[esi+Node.shared],0
		jnz	.shared
		
		; The node has only one possible attribute:
		; Create the attribute
		
		call	createElement
		jnz	.error		
		add	ebx,1	; set bit 0 to 1
		jmp	.next
		
		; The node has multiple possible attributes:
		; Create a submap and point to it
.shared:	push	esi
		mov	ebx,esi
		mov	esi,[esi+Node.succ]
		call	createElementMap
		pop	esi
		jnz	.error

 		; Point to the attribute or map (depending on bit 0)
.next:		movzx	eax,[esi+Node.char]
		mov	[edi+eax*4],ebx
		
		mov	esi,[esi+Node.next]
		or	esi,esi
		jnz	.lp
		
		mov	ebx,edi
.ok:		xor	eax,eax
.error:		pop	edi
		pop	esi
		ret

.empty:		mov	ebx,noElement
		jmp	.ok
;[c]
;[of]:createElement
;[c]Create an element
;[c]
;[c]ARGUMENTS
;[c]	esi
;[c]		Node
;[c]
;[c]RETURN VALUES
;[c]	zf+eax
;[c]		Error code
;[c]	ebx
;[c]		Element
;[c]
createElement:

		; Compute the number of remaining chars
		xor	eax,eax
		mov	ebx,esi
		sub	eax,1			; avoid jmp+1
.nextChar:	add	eax,1
		mov	ebx,[ebx+Node.succ]
		or	ebx,ebx
		jnz	.nextChar
		
		; Add null char
		; Add flag char
		; Add pointer
		; Align to multiple of 4
		add	eax,1+1+3+4
		and	eax,not 3
		
		cmp	[esi+Node.collectionClass],0
		jnz	.coll
		add	eax,TextElementClass.sizeof - 4

		; Allocate memory for element
.coll:		mov	ecx,4
		push	edi
		mov	edx,[ebp+TreeBuilder.context]
		call	allocateMemory
		mov	ebx,edi
		pop	edi
		jnz	.error
		
		; Copy remaining chars into attribute
		mov	ecx,ebx
		mov	eax,[esi+Node.succ]
		mov	edx,esi
		or	eax,eax
		jz	.skip2
.nextChar2:	mov	edx,eax
		mov	al,[edx+Node.char]
		mov	[ecx],al
		add	ecx,1
		mov	eax,[edx+Node.succ]
		or	eax,eax
		jnz	.nextChar2
.skip2:		mov	byte [ecx],0
		add	ecx,1

		mov	eax,[edx+Node.collectionClass]
		or	eax,eax
		setz	byte [ecx]
		add	ecx,1+3
		and	ecx,not 3
		or	eax,eax
		jz	.txt
		
		; Write Collection Class Information
		mov	[ecx],eax
		jmp	.ok

		; Write Text Class Information
.txt:		push	esi
		push	edi
		mov	eax,[edx+Node.index]
		mov	esi,[edx+Node.nameBegin]
		mov	edi,[edx+Node.nameLimit]
		mov	[ecx+TextElementClass.offset],eax
		mov	[ecx+TextElementClass.begin],esi
		mov	[ecx+TextElementClass.limit],edi
		pop	edi
		pop	esi

.ok:		xor	eax,eax
.error:		ret
;[cf]
;[of]:createTerminalElement
;[c]Create a terminal element
;[c]
;[c]ARGUMENTS
;[c]	esi
;[c]		Node
;[c]
;[c]RETURN VALUES
;[c]	zf+eax
;[c]		Error code
;[c]	ebx
;[c]		Element
;[c]
createTerminalElement:

		; Add null char
		; Add flag char
		; Add pointer
		; Align to multiple of 4
		mov	eax,8
		cmp	[esi+Node.collectionClass],0
		jnz	.coll
		add	eax,TextElementClass.sizeof - 4

		; Allocate memory for element
.coll:		mov	ecx,4
		push	edi
		mov	edx,[ebp+TreeBuilder.context]
		call	allocateMemory
		mov	ebx,edi
		pop	edi
		jnz	.error
		
		; Copy remaining chars into attribute
		mov	ecx,ebx
		mov	byte [ecx],0
		add	ecx,1
		mov	eax,[esi+Node.collectionClass]
		or	eax,eax
		setz	byte [ecx]
		add	ecx,1+3
		and	ecx,not 3
		or	eax,eax
		jz	.txt
		
		; Write Collection Class Information
		mov	[ecx],eax
		jmp	.ok

		; Write Text Class Information
.txt:		push	edi
		mov	eax,[esi+Node.index]
		mov	edx,[esi+Node.nameBegin]
		mov	edi,[esi+Node.nameLimit]
		mov	[ecx+TextElementClass.offset],eax
		mov	[ecx+TextElementClass.begin],edx
		mov	[ecx+TextElementClass.limit],edi
		pop	edi

.ok:		xor	eax,eax
.error:		ret
;[cf]
;[cf]
;[of]:	createAttributeMap
;[c]Create an attribute map for one node
;[c]
;[c]ARGUMENTS
;[c]	ebx
;[c]		-1 if the node does not have an attribute
;[c]	esi
;[c]		the node
;[c]	ebp
;[c]	
;[c]RETURN VALUE
;[c]	eax+zf
;[c]		error code
;[c]	ebx
;[c]		the map
;[c]
createAttributeMap:
		; The node is null: return an empty map
		or	esi,esi
		jnz	.noEmpty
		mov	ebx,noAttribute		
		xor	eax,eax
		ret
.noEmpty:
		
		push	esi
		push	edi

		call	createMap
		
		; If the previous node contains an offset, create the
		; attribute and make any non-name char a
		; target to this attribute.
		;
		; Note: it cannot happen on the root node since
		; a name cannot be empty.
		cmp	ebx,-1
		jz	.noAttr
		
		push	edi
		xor	eax,eax
		call	allocateAttribute
		pop	edx
		jnz	.error
		mov	byte [edi],0
		mov	[edi+1],ebx
		mov	ebx,edx
		add	edi,3	; set bit 0 & 1 to 1
		call	mapTerminal
		mov	edi,ebx
.noAttr:

		; Process each node
.lp:		cmp	[esi+Node.shared],0
		jnz	.shared
		
		; The node has only one possible attribute:
		; Create the attribute
		
		call	createAttribute
		jnz	.error		
		add	ebx,1	; set bit 0 to 1
		jmp	.next
		
		; The node has multiple possible attributes:
		; Create a submap and point to it
.shared:	push	esi
		mov	ebx,[esi+Node.index]
		mov	esi,[esi+Node.succ]
		call	createAttributeMap
		pop	esi
		jnz	.error

 		; Point to the attribute or map (depending on bit 0)
.next:		movzx	eax,[esi+Node.char]
		mov	[edi+eax*4],ebx
		
		mov	esi,[esi+Node.next]
		or	esi,esi
		jnz	.lp
		
		xor	eax,eax
		mov	ebx,edi
.error:		pop	edi
		pop	esi
		ret
;[c]
;[of]:mapTerminal
;[c]Set the object for each invalid name char
;[c]
;[c]ARGUMENT
;[c]	edx
;[c]		the map
;[c]	edi
;[c]		the object
;[c]
mapTerminal:	push	esi

		mov	ecx,256
		lea	esi,[terminalChar]
.next:		cmp	byte [esi],1
		jz	.skip
		mov	[edx],edi
.skip:		add	edx,4
		add	esi,1
		sub	ecx,1
		jnz	.next
		
		pop	esi
		ret
;[cf]
;[of]:createMap
createMap:
		; Allocate a map
		call	allocateMap
		jnz	.error

		; Initialize map
		push	edi
		mov	ecx,256
		mov	eax,1
		rep
		stosd
		pop	edi
		
		xor	eax,eax
.error:		ret
;[cf]
;[of]:allocateMap
;[c]Allocates a map
;[c]
;[c]ARGUMENTS
;[c]	ebp
;[c]		the tree builder local variables
;[c]		
;[c]RETURN VALUES
;[c]	zf
;[c]		0 if not  enough memory
;[c]	eax
;[c]		error code
;[c]	edi
;[c]		the map
;[c]
MAP_SIZE	equ	256*4
MAP_ALIGN	equ	64
;[c]
allocateMap:	mov	ecx,MAP_ALIGN
		mov	eax,MAP_SIZE
		mov	edx,[ebp+TreeBuilder.context]
		jmp	allocateMemory
;[cf]
;[of]:createAttribute
;[c]Create an attribute
;[c]
;[c]ARGUMENTS
;[c]	esi
;[c]		Node
;[c]
;[c]RETURN VALUES
;[c]	zf+eax
;[c]		Error code
;[c]	ebx
;[c]		Attribute
;[c]
createAttribute:

		; Compute the number of remaining chars
		xor	eax,eax
		mov	ebx,esi
		sub	eax,1			; avoid jmp+1
.nextChar:	add	eax,1
		mov	ebx,[ebx+Node.succ]
		or	ebx,ebx
		jnz	.nextChar
		
		; Allocate memory for attribute
		push	edi
		call	allocateAttribute
		mov	ebx,edi
		pop	edi
		jnz	.error

		; Copy remaining chars into attribute
		mov	ecx,ebx
		mov	eax,[esi+Node.succ]
		or	eax,eax
		jz	.skip2
.nextChar2:	mov	edx,eax
		mov	al,[edx+Node.char]
		mov	[ecx],al
		add	ecx,1
		mov	eax,[edx+Node.succ]
		or	eax,eax
		jnz	.nextChar2
.skip2:		mov	byte [ecx],0

		; Append the index from the last node (edx)
		mov	eax,[edx+Node.index]
		mov	[ecx+1],eax
		
		xor	eax,eax
.error:		ret
;[cf]
;[of]:allocateAttribute
;[c]Allocates an attribute
;[c]
;[c]ARGUMENTS
;[c]	eax
;[c]		extra size
;[c]	ebp
;[c]		the tree builder local variables
;[c]		
;[c]RETURN VALUES
;[c]	zf
;[c]		0 if not  enough memory
;[c]	eax
;[c]		error code
;[c]	edi
;[c]		the map
;[c]
allocateAttribute:
		mov	ecx,4
		add	eax,1+4
		mov	edx,[ebp+TreeBuilder.context]
		jmp	allocateMemory
;[cf]
;[cf]
;[c]
;[c]CLASSES
;[of]:	Node
;[of]:Comment
;[c]Node
;[c]
;[c]A Node object is just a part of a tree that is a temporary object used to
;[c]build the maps of attributes/elements.
;[c]
;[c]The goal is to have an efficient implementation of a dictionary,
;[c]i.e. a set of name -> value.
;[c]
;[c]A null node pointer is an empty node. So a list is simply initialized
;[c]by clearing a register.
;[c]
;[c]A node is also a list if node, one for each char that can appear at this 
;[c]level. The list of nodes is obtained by following the 'next' property.
;[c]
;[c]When a character matches, the 'succ' gives the next node for the next
;[c]character.
;[c]
;       succ
; Node ---------> Node
;  |               |next
;  |next
;  |
;  |    succ
; Node ---------> Node
;  |
;  |next
;  |
;  |    succ
; Node ---------> Node
;  |
;  |next
;  .
;  .

;[cf]
;[of]:Definitions
		struct		Node
.next		rd	1
.succ		rd	1
.char		rb	1
.shared		rb	1
		rb	2
.index		rd	1
.collectionClass	rd	1
.nameBegin	rd	1
.nameLimit	rd	1
		ends
;[cf]
;[c]
;[of]:findOrCreateNode
;[c]Find or create a node
;[c]
;[c]ARGUMENTS
;[c]	esi
;[c]		the parent node or nil
;[c]	al
;[c]		the char
;[c]
;[c]RETURN VALUES
;[c]	zf
;[c]		0 if error
;[c]	eax
;[c]		error code
;[c]	esi
;[c]		the new parent node
;[c]	edi
;[c]		the found or created node
;[c]
;[c]REMARKS
;[c]	If the node already exists, the shared flag is set.
;[c]	
findOrCreateNode:

		call	findNode
		jz	.notFound
		
		mov	[edi+Node.shared],1
		jmp	.ok

		; Allocate a node
.notFound:	push	eax
		call	allocateNode
		pop	edx
		jnz	.error

		; Append node
		call	appendNode
		
		; Initialize node
		xor	eax,eax
		mov	[edi+Node.next],eax
		mov	[edi+Node.succ],eax
		mov	[edi+Node.char],dl
		mov	[edi+Node.shared],al
		mov	[edi+Node.index],-1
		mov	[edi+Node.collectionClass],eax
		
.ok:		xor	eax,eax
.error:		ret
;[cf]
;[of]:allocateNode
;[c]Allocates a node
;[c]
;[c]ARGUMENTS
;[c]	ebp
;[c]		the tree builder local variables
;[c]		
;[c]RETURN VALUES
;[c]	zf
;[c]		0 if not  enough memory
;[c]	eax
;[c]		error code
;[c]	edi
;[c]		the node
;[c]
allocateNode:	mov	ecx,4
		mov	eax,Node.sizeof
		mov	edx,[ebp+TreeBuilder.context]
		jmp	allocateMemory
;[cf]
;[of]:appendNode
;[c]Append a node
;[c]
;[c]ARGUMENTS
;[c]	esi
;[c]		parent node
;[c]	edi
;[c]		node to append to parent
;[c]
;[c]RETURN VALUES
;[c]	esi
;[c]		parent node
;[c]	edi
;[c]		node to append
;[c]
;[c]REMARKS
;[c]	if the parent node is nil (i.e. empty), edi becomes the parent.
;[c]
appendNode:	call	lastNode
		jz	.empty
		mov	[esi+Node.next],edi
		ret
.empty:		mov	esi,edi
		ret
;[cf]
;[of]:findNode
;[c]Find a node
;[c]
;[c]ARGUMENTS
;[c]	al
;[c]		char
;[c]	esi
;[c]		parent node
;[c]
;[c]RETURN VALUES
;[c]	edi
;[c]		search node or nil
;[c]	zf
;[c]		1 if not found
;[c]	
findNode:	mov	edi,esi
		jmp	.first

.next:		cmp	al,[edi+Node.char]
		jz	.done
		mov	edi,[edi+Node.next]
.first:		or	edi,edi
		jnz	.next
	
.done:		or	edi,edi
		ret
;[cf]
;[of]:lastNode
;[c]Returns the last node
;[c]
;[c]ARGUMENTS
;[c]	esi
;[c]		parent node
;[c]
;[c]RETURN VALUES
;[c]	esi
;[c]		the last node
;[c]	zf
;[c]		1 if the list of nodes is empty
;[c]	
lastNode:	or	esi,esi
		jnz	.first
		ret
.next:		mov	esi,eax
.first:		mov	eax,[esi+Node.next]
		or	eax,eax
		jnz	.next
		or	esi,esi
		ret
;[cf]
;[cf]
;[cf]
;[cf]
;[cf]
;[c]
;[of]:Memory
;[of]:allocateMemory
;[c]Allocates a memory block
;[c]
;[c]ARGUMENTS
;[c]	eax
;[c]		size
;[c]	ecx
;[c]		alignment
;[c]	edx
;[c]		context
;[c]		
;[c]RETURN VALUES
;[c]	eax+zf
;[c]		error code
;[c]	edi
;[c]		the object
;[c]
allocateMemory:	mov	edi,[edx+MemoryContext.base]
		sub	ecx,1
		add	edi,ecx
		not	ecx
		and	edi,ecx
		add	eax,edi
		cmp	eax,[edx+MemoryContext.limit]
		jnc	.chunkFull
		mov	[edx+MemoryContext.base],eax
		xor	eax,eax
		ret
;[c]
.chunkFull:	push	edx
		push	ebp
		mov	ebp,edx
		sub	eax,edi
		call	extendMemory	; let's assume that alignment is ok
		pop	ebp
		pop	edx
		ret
;[cf]
;[of]:extendMemory
;[c]Extend memory
;[c]
;[c]When there is not enough memory available, a new chunk is created
;[c]
;[c]ARGUMENTS
;[c]	eax
;[c]		size
;[c]	ebp
;[c]		the parse context
;[c]
;[c]RETURN VALUES
;[c]	eax+zf
;[c]		error code
;[c]	edi
;[c]		the allocated memory buffer
;[c]
extendMemory:	push	eax
		mov	ecx,[ebp+MemoryContext.chunkSize]
		cmp	eax,ecx
		jc	.a
		add	eax,ecx
.a:		call	newChunk
		pop	ecx
		jnz	.error
		
		mov	edi,[ebp+MemoryContext.base]
		add	ecx,edi
		mov	[ebp+MemoryContext.base],ecx

		xor	eax,eax
.error:		ret
;[cf]
;[of]:releaseChunks
;[c]Release the memory handled by the memory context
;[c]
;[c]ARGUMENTS
;[c]	ebp
;[c]		the context
;[c]
releaseChunks:	push	ebx
		mov	eax,[ebp+MemoryContext.chunks]
		or	eax,eax
		jz	.done
.loop:		mov	ebx,[eax]
		call	freeChunk
		mov	eax,ebx
		or	eax,eax
		jnz	.loop
.done:		pop	ebx
		ret		
;[cf]
;[of]:newChunk
;[c]Create a new chunk of memory
;[c]
;[c]ARGUMENTS
;[c]	ebp
;[c]		the parse context
;[c]	eax
;[c]		the size of chunk
;[c]
;[c]RETURN VALUES
;[c]	eax+zf
;[c]		error code
;[c]
newChunk:	mov	ecx,eax
		add	eax,HEADER_SIZE		; size for link
		push	ecx
		push	eax
		call	[mallocFunction]
		add	esp,4
		pop	ecx
		or	eax,eax
		jz	.outOfMemory
		mov	edx,[ebp+MemoryContext.chunks]
		mov	[eax],edx
		mov	[ebp+MemoryContext.chunks],eax
		add	eax,HEADER_SIZE
		add	ecx,eax
		mov	[ebp+MemoryContext.base],eax
		mov	[ebp+MemoryContext.limit],ecx
		
		xor	eax,eax
		ret
;[c]
.outOfMemory:	mov	eax,RC_MEMORY
		or	eax,eax
		ret
;[cf]
;[of]:freeChunk
;[c]Free a chunk
;[c]
;[c]ARGUMENTS
;[c]	eax
;[c]		the chunk
;[c]
freeChunk:	push	eax
		call	[freeFunction]
		add	esp,4
		ret
;[cf]
;[cf]
;[c]
;[of]:BSS
mallocFunction	rd	1
freeFunction	rd	1

;[cf]
;[of]:Data
		align	4

;[of]:latin1 maps
;[c]Latin1 valid characters - unused
;[c]
;[c]latin1NameChar1:
;[c]		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
;[c]		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
;[c]		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
;[c]		db	0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0 ; :
;[c]		db	0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
;[c]		db	1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1 ; _
;[c]		db	0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
;[c]		db	1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0
;[c]		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
;[c]		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
;[c]		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
;[c]		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
;[c]		db	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
;[c]		db	1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,1
;[c]		db	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
;[c]		db	1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,1
;[c]
;[c]latin1NameChar:
;[c]		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
;[c]		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
;[c]		db	0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0	; -.
;[c]		db	1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0 ; :
;[c]		db	0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
;[c]		db	1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1 ; _
;[c]		db	0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
;[c]		db	1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0
;[c]		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
;[c]		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
;[c]		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
;[c]		db	0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0
;[c]		db	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
;[c]		db	1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,1
;[c]		db	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
;[c]		db	1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,1
;[cf]
;[of]:S
;[c]The blank map
;[c]
;[c]	Flag for each character:
;[c]	- 1 for blank characters (9, 10, 13, 32)
;[c]	- 0 for other characters
;[c]
S:		db	0,0,0,0,0,0,0,0,0,1,1,0,0,1,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
;[cf]
;[of]:terminalChar
;[c]Any zero char is a terminator for an attribute name or an element name
;[c]
terminalChar:	db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0	; -.
		db	1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0 ; :
		db	0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		db	1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1 ; _
		db	0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		db	1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0
		db	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		db	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		db	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		db	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		db	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		db	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		db	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		db	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
;[cf]
;[of]:digit
;[c]Two flags for each chars:
;[c]	- bit 0 for decimal chars
;[c]	- bit 1 for hexadecimal chars
;[c]
digit:		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	3,3,3,3,3,3,3,3,3,3,0,0,0,0,0,0
		db	0,2,2,2,2,2,2,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,2,2,2,2,2,2,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
;[cf]
;[of]:xdigit
;[c]Hexadecimal value of a char, assuming that the char is a valid 
;[c]hexadecimal digit.
;[c]
xdigit:		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,1,2,3,4,5,6,7,8,9,0,0,0,0,0,0
		db	0,10,11,12,13,14,15,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,10,11,12,13,14,15,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
;[cf]
;[of]:noElement / noAttribute
;[c]The default map for element without child elements (attribute only elements).
;[c]
noAttribute:
noElement:
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
;[cf]
;[c]
;[of]:Document Class
;[c]Document Class
;[c]
DocumentClass:	dd	documentName
		dd	documentName+8
		dd	Element.sizeof + 1 * Attribute.sizeof
		dd	1
		dd	0
		dd	nameAttr
		dd	documentElem

;[c]
;[of]:Attribute Class
;[c]Attribute Class
;[c]
AttributeClass:	dd	attributeName
		dd	attributeName+9
		dd	Element.sizeof + 1 * Attribute.sizeof
		dd	1
		dd	0
		dd	nameAttr
		dd	noElement

;[cf]
;[of]:Text Class
;[c]Attribute Class
;[c]
TextClass:	dd	textName
		dd	textName+4
		dd	Element.sizeof + 1 * Attribute.sizeof
		dd	1
		dd	0
		dd	nameAttr
		dd	noElement

;[cf]
;[of]:Collection Class
CollectionClass:
		dd	collectionName
		dd	collectionName+10
		dd	Element.sizeof + 1 * Attribute.sizeof
		dd	1
		dd	0
		dd	nameAttr
		dd	documentElem
;[cf]
;[c]
;[of]:Map for root and collection elements
		align	4
documentElem:	dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,.a+1,1,.c+1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,.t+1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1

		align	2
.a:		db	"ttribute",0,0
		align	4
		dd	AttributeClass

		align	2
.t:		db	"ext",0,0
		align	4
		dd	TextClass

.c:		align	2
		db	"ollection",0,0
		align	4
		dd	CollectionClass
;[cf]
;[of]:Map for 'name' attribute

		align	4
nameAttr:	dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,.name+1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
		dd	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1

		align	2
.name:		db	"ame",0
		dd	Element.sizeof + 0 * Attribute.sizeof

;[cf]
;[c]
;[of]:Element Names
documentName	db	"document",0
attributeName	db	"attribute",0
textName	db	"text",0
collectionName	db	"collection",0
;[cf]
;[cf]
;[cf]
